// Package p2p including peer discovery using discv5, gossip-sub
// using libp2p, and handing peer lifecycles + handshakes.
package p2p

import (
	"context"
	"crypto/ecdsa"
	"fmt"
	"net"
	"strings"
	"sync"
	"time"

	"witness/p2p/peers"

	"github.com/libp2p/go-libp2p"
	"github.com/multiformats/go-multiaddr"
	"github.com/prysmaticlabs/prysm/v3/async/event"

	"github.com/ethereum/go-ethereum/p2p/enode"
	"github.com/ethereum/go-ethereum/p2p/enr"
	pubsub "github.com/libp2p/go-libp2p-pubsub"
	"github.com/libp2p/go-libp2p/core/host"
	"github.com/libp2p/go-libp2p/core/network"
	"github.com/libp2p/go-libp2p/core/peer"
	"github.com/libp2p/go-libp2p/core/protocol"
	"github.com/pkg/errors"
	prysmnetwork "github.com/prysmaticlabs/prysm/v3/network"
)

// In the event that we are at our peer limit, we
// stop looking for new peers and instead poll
// for the current peer limit status for the time period
// defined below.
var pollingPeriod = time.Second * 12

// maxDialTimeout is the timeout for a single peer dial.
var maxDialTimeout = time.Second * 10

// pubsubQueueSize is the size that we assign to our validation queue and outbound message queue for
// gossipsub.
const pubsubQueueSize = 80

const GossipMaxSize = 100 * 1 << 20 // 10MB

const MaxChunkSize = 10 * 1 << 20 // 1MB

type subSession struct {
	sub       event.Subscription
	listeners int
}

// Service for managing peer to peer (p2p) networking.
type Service struct {
	started          bool
	cancel           context.CancelFunc
	cfg              *Config
	privKey          *ecdsa.PrivateKey
	pubsub           *pubsub.PubSub
	joinedTopics     map[string]*pubsub.Topic
	joinedTopicsLock sync.Mutex
	dv5Listener      Listener
	startupErr       error
	ctx              context.Context
	host             host.Host
	peers            *peers.Status
	subManage        map[string]*subSession
	subManageLock    sync.Mutex
}

// NewService initializes a new p2p service compatible with shared.Service interface. No
// connections are made until the Start function is called during the service registry startup.
func NewService(ctx context.Context, cfg *Config) (*Service, error) {
	var err error
	ctx, cancel := context.WithCancel(ctx)
	_ = cancel // govet fix for lost cancel. Cancel is handled in service.Stop().
	s := &Service{
		ctx:          ctx,
		cancel:       cancel,
		cfg:          cfg,
		subManage:    make(map[string]*subSession),
		joinedTopics: make(map[string]*pubsub.Topic),
	}

	dv5Nodes, _ := parseBootStrapAddrs(s.cfg.BootstrapNodeAddr)
	cfg.Discv5BootStrapAddr = dv5Nodes

	s.privKey, err = privKey(s.cfg)
	if err != nil {
		log.WithError(err).Error("Failed to generate p2p private key")
		return nil, err
	}
	// get external ip address
	ipAddr := prysmnetwork.IPAddr()
	if cfg.HostAddress != "" {
		ipAddr = net.ParseIP("0.0.0.0")
		if ipAddr.To4() == nil && ipAddr.To16() == nil {
			log.Errorf("Invalid host address given: %s", ipAddr.String())
			return nil, fmt.Errorf("invalid host address given: %s", ipAddr.String())
		}
	}

	opts := s.buildOptions(ipAddr, s.privKey)
	h, err := libp2p.New(opts...)
	if err != nil {
		log.WithError(err).Error("Failed to create p2p host")
		return nil, err
	}

	s.host = h
	// Gossipsub registration is done before we add in any new peers
	// due to libp2p's gossipsub implementation not taking into
	// account previously added peers when creating the gossipsub
	// object.
	psOpts := s.pubsubOptions()
	// Set the pubsub global parameters that we require.
	setPubSubParameters()
	// Reinitialize them in the event we are running a custom config.
	gs, err := pubsub.NewGossipSub(s.ctx, s.host, psOpts...)
	if err != nil {
		log.WithError(err).Error("Failed to start pubsub")
		return nil, err
	}

	s.pubsub = gs
	s.peers = peers.NewStatus(s.ctx, &peers.StatusConfig{
		MaxInboundPeers:  int(cfg.MaxInboundPeers),
		MaxOutboundPeers: int(cfg.MaxOutboundPeers),
		MaxPeers:         int(cfg.MaxPeers),
		MaxBadResponses:  3,
	})

	return s, nil
}

// Start the p2p service.
func (s *Service) Start() {
	if s.started {
		log.Error("Attempted to start p2p service when it was already started")
		return
	}

	if s.cfg.UseDiscovery {
		ipAddr := prysmnetwork.IPAddr()
		listener, err := s.startDiscoveryV5(
			ipAddr,
			s.privKey,
		)
		if err != nil {
			log.WithError(err).Fatal("Failed to start discovery")
			s.startupErr = err
			return
		}
		log.Debug("debug: connecting to bootnode ", len(s.cfg.Discv5BootStrapAddr))
		err = s.connectToBootnodes()
		if err != nil {
			log.WithError(err).Error("Could not add bootnode to the exclusion list")
			s.startupErr = err
			return
		}
		s.dv5Listener = listener
		go s.listenForNewNodes()
	}

	s.started = true

	if len(s.cfg.StaticPeers) > 0 {
		addrs, err := PeersFromStringAddrs(s.cfg.StaticPeers)
		if err != nil {
			log.WithError(err).Error("Could not connect to static peer")
		}
		s.connectWithAllPeers(addrs, false)
	}

	multiAddrs := s.host.Network().ListenAddresses()
	logIPAddr(s.host.ID(), multiAddrs...)

	p2pHostAddress := s.cfg.HostAddress
	p2pTCPPort := s.cfg.TCPPort

	if p2pHostAddress != "" {
		logExternalIPAddr(s.host.ID(), p2pHostAddress, p2pTCPPort)
		verifyConnectivity(p2pHostAddress, p2pTCPPort, "tcp")
	}

}

// Stop the p2p service and terminate all peer connections.
func (s *Service) Stop() error {
	defer s.cancel()
	s.started = false
	if s.dv5Listener != nil {
		s.dv5Listener.Close()
	}
	return nil
}

// Status of the p2p service. Will return an error if the service is considered unhealthy to
// indicate that this node should not serve traffic until the issue has been resolved.
func (s *Service) Status() error {
	if !s.started {
		return errors.New("not running")
	}
	if s.startupErr != nil {
		return s.startupErr
	}
	return nil
}

// Started returns true if the p2p service has successfully started.
func (s *Service) Started() bool {
	return s.started
}

// PubSub returns the p2p pubsub framework.
func (s *Service) PubSub() *pubsub.PubSub {
	return s.pubsub
}

// Host returns the currently running libp2p
// host of the service.
func (s *Service) Host() host.Host {
	return s.host
}

// SetStreamHandler sets the protocol handler on the p2p host multiplexer.
// This method is a pass through to libp2pcore.Host.SetStreamHandler.
func (s *Service) SetStreamHandler(topic string, handler network.StreamHandler) {
	s.host.SetStreamHandler(protocol.ID(topic), handler)
}

// PeerID returns the Peer ID of the local peer.
func (s *Service) PeerID() peer.ID {
	return s.host.ID()
}

// Disconnect from a peer.
func (s *Service) Disconnect(pid peer.ID) error {
	return s.host.Network().ClosePeer(pid)
}

// Connect to a specific peer.
func (s *Service) Connect(pi peer.AddrInfo) error {
	return s.host.Connect(s.ctx, pi)
}

func (s *Service) connectWithAllPeers(multiAddrs []multiaddr.Multiaddr, isBootnode bool) {
	addrInfos, err := peer.AddrInfosFromP2pAddrs(multiAddrs...)
	if err != nil {
		log.WithError(err).Error("Could not convert to peer address info's from multiaddresses")
		return
	}
	for _, info := range addrInfos {
		for _, addr := range info.Addrs {
			ipAddr := strings.Split(addr.String(), "/")[2]
			if s.cfg.IgnoreLocalIP && s.isLocalIP(ipAddr) {
				continue
			}
		}

		// make each dial non-blocking
		go func(info peer.AddrInfo) {
			log.Debugf("connecting to peer %s", info.String())
			if err := s.connectWithPeer(s.ctx, info); err != nil {
				log.WithError(err).Errorf("Could not connect with peer %s, err %s", info.String(), err.Error())
			} else if isBootnode {
				// save to file
			}
		}(info)
	}
}

func (s *Service) connectWithPeer(ctx context.Context, info peer.AddrInfo) error {
	if info.ID == s.host.ID() {
		return nil
	}
	if s.peers.IsBad(info.ID) {
		return errors.New("refused to connect to bad peer")
	}

	for _, addr := range info.Addrs {
		ipAddr := strings.Split(addr.String(), "/")[2]
		if s.cfg.IgnoreLocalIP && s.isLocalIP(ipAddr) {
			return nil
		}
	}

	ctx, cancel := context.WithTimeout(ctx, maxDialTimeout)
	defer cancel()
	if err := s.host.Connect(ctx, info); err != nil {
		s.peers.IncBadResponses(info.ID)
		return err
	}
	log.WithField("peer", info.ID).Debug("Connected to new peer")
	return nil
}

func (s *Service) connectToBootnodes() error {
	nodes := make([]*enode.Node, 0, len(s.cfg.Discv5BootStrapAddr))
	for _, addr := range s.cfg.Discv5BootStrapAddr {
		bootNode, err := enode.Parse(enode.ValidSchemes, addr)
		if err != nil {
			return err
		}
		log.Debug("Adding bootnode to the list", bootNode.String())
		// do not dial bootnodes with their tcp ports not set
		if err := bootNode.Record().Load(enr.WithEntry("tcp", new(enr.TCP))); err != nil {
			if !enr.IsNotFound(err) {
				log.WithError(err).Error("Could not retrieve tcp port")
			}
			continue
		}

		if s.cfg.IgnoreLocalIP && s.isLocalIP(bootNode.IP().String()) {
			continue
		}

		nodes = append(nodes, bootNode)
	}
	multiAddresses := convertToMultiAddr(nodes)
	s.connectWithAllPeers(multiAddresses, true)
	return nil
}

func (s *Service) isLocalIP(ipAddr string) bool {
	return ipAddr == s.cfg.LocalIP || ipAddr == s.cfg.HostAddress || ipAddr == "127.0.0.1"
}
