package p2p

import (
	"crypto/ecdsa"
	"crypto/rand"
	"encoding/hex"
	"fmt"
	"net"
	"os"
	"time"

	"github.com/libp2p/go-libp2p/core/crypto"
	"github.com/pkg/errors"
	ecdsaprysm "github.com/prysmaticlabs/prysm/v3/crypto/ecdsa"
	"github.com/sirupsen/logrus"
)

const dialTimeout = 1 * time.Second

const enrRecordFile = "discovery-enr"

const peerIDFile = "peer-id"

// Determines a private key for p2p networking from the p2p service's
// configuration struct. If no key is found, it generates a new one.
func privKey(cfg *Config) (*ecdsa.PrivateKey, error) {
	privateKeyPath := cfg.PrivateKeyPath
	_, err := os.Stat(privateKeyPath)
	privateKeyExist := !os.IsNotExist(err)
	if err != nil && privateKeyExist {
		return nil, err
	}
	if privateKeyPath == "" || !privateKeyExist {
		priv, _, err := crypto.GenerateSecp256k1Key(rand.Reader)
		if err != nil {
			return nil, err
		}
		return ecdsaprysm.ConvertFromInterfacePrivKey(priv)
	}
	return privKeyFromFile(privateKeyPath)
}

// Retrieves a p2p networking private key from a file path.
func privKeyFromFile(path string) (*ecdsa.PrivateKey, error) {
	src, err := os.ReadFile(path) // #nosec G304
	if err != nil {
		log.WithError(err).Error("Error reading private key from file")
		return nil, err
	}
	dst := make([]byte, hex.DecodedLen(len(src)))
	_, err = hex.Decode(dst, src)
	if err != nil {
		return nil, errors.Wrap(err, "failed to decode hex string")
	}
	unmarshalledKey, err := crypto.UnmarshalSecp256k1PrivateKey(dst)
	if err != nil {
		return nil, err
	}
	return ecdsaprysm.ConvertFromInterfacePrivKey(unmarshalledKey)
}

// Attempt to dial an address to verify its connectivity
func verifyConnectivity(addr string, port uint, protocol string) {
	if addr != "" {
		a := net.JoinHostPort(addr, fmt.Sprintf("%d", port))
		fields := logrus.Fields{
			"protocol": protocol,
			"address":  a,
		}
		conn, err := net.DialTimeout(protocol, a, dialTimeout)
		if err != nil {
			log.WithError(err).WithFields(fields).Warn("IP address is not accessible")
			return
		}
		if err := conn.Close(); err != nil {
			log.WithError(err).Debug("Could not close connection")
		}
	}
}

func enrRecordToFile(enr string) {
	// create or overwrite
	f, err := os.Create(enrRecordFile)
	if err != nil {
		log.WithError(err).Warn("Error creating enrRecordFile")
		return
	}
	defer f.Close()
	_, err = f.WriteString(enr)
	if err != nil {
		log.WithError(err).Warn("Error writing enrRecordFile")
		return
	}
}

func myPeerIDToFile(peerID string) {
	f, err := os.Create(peerIDFile)
	if err != nil {
		log.WithError(err).Warn("Error creating peerIDFile")
		return
	}
	defer f.Close()
	_, err = f.WriteString(peerID)
	if err != nil {
		log.WithError(err).Warn("Error writing peerIDFile")
		return
	}
}
